import cv2

from aperturedb import ParallelLoader
from aperturedb import CSVParser

HEADER_PATH = "filename"
PROPERTIES  = "properties"
CONSTRAINTS = "constraints"


class VideoGeneratorCSV(CSVParser.CSVParser):
    """
    **ApertureDB Video Data loader.**

    .. warning::
        Deprecated. Use :class:`~aperturedb.VideoDataCSV.VideoDataCSV` instead.


    .. important::

        Expects a csv file with the following columns:

            ``filename``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``

    Example csv file::

        filename,id,label,constaint_id
        /home/user/file1.jpg,321423532,dog,321423532
        /home/user/file2.jpg,42342522,cat,4234252
        ...


    """

    def __init__(self, filename, check_video=True):

        super().__init__(filename)

        self.check_video = check_video

        self.props_keys       = [x for x in self.header[1:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.constraints_keys = [x for x in self.header[1:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]

    def getitem(self, idx):

        filename   = self.df.loc[idx, HEADER_PATH]
        data = {}

        video_ok, video = self.load_video(filename)
        if not video_ok:
            print("Error loading video: " + filename)
            raise Exception("Error loading video: " + filename)

        data["video_blob"] = video

        properties  = self.parse_properties(self.df, idx)
        constraints = self.parse_constraints(self.df, idx)

        if properties:
            data[PROPERTIES] = properties

        if constraints:
            data[CONSTRAINTS] = constraints

        return data

    def load_video(self, filename):

        if self.check_video:
            try:
                a = cv2.VideoCapture(filename)
                if a.isOpened() == False:
                    print("Video reading Error:", filename)
            except:
                print("Video Error:", filename)

        try:
            fd = open(filename, "rb")
            buff = fd.read()
            fd.close()
            return True, buff
        except:
            print("Video Error:", filename)

        return False, None

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] != HEADER_PATH:
            raise Exception(
                "Error with CSV file field: filename. Must be first field")


class VideoLoader(ParallelLoader.ParallelLoader):
    """**ApertureDB Video Loader.**

    This class is to be used in combination with a "generator",
    for example :class:`~aperturedb.VideoLoader.VideoGeneratorCSV`.
    The generator must be an iterable object that generated `video_data`
    elements.

    Example::

            image_data = {
                "properties":  properties,
                "constraints": constraints,
                "operations":  operations,
                "video_blob":    (bytes),
            }
    """

    def __init__(self, db, dry_run=False):

        super().__init__(db, dry_run=dry_run)

        self.type = "video"

    def generate_batch(self, video_data):

        q = []
        blobs = []

        for data in video_data:

            ai = {
                "AddVideo": {
                }
            }

            if "properties" in data:
                ai["AddVideo"]["properties"] = data["properties"]
            if "constraints" in data:
                ai["AddVideo"]["if_not_found"] = data["constraints"]
            if "operations" in data:
                ai["AddVideo"]["operations"] = data["operations"]
            if "format" in data:
                ai["AddVideo"]["format"] = data["format"]

            if "video_blob" not in data or len(data["video_blob"]) == 0:
                print("WARNING: Skipping empty video.")
                continue

            blobs.append(data["video_blob"])
            q.append(ai)

        return q, blobs



import pandas as pd
import logging

logger = logging.getLogger(__name__)

ENTITY_CLASS      = "EntityClass"
CONTRAINTS_PREFIX = "constraint_"
PROPERTIES  = "properties"
CONSTRAINTS = "constraints"


class CSVParser():
    """**ApertureDB General CSV Parser for Loaders.**
    ...
    """

    def __init__(self, filename):

        self.df = pd.read_csv(filename)

        self.validate()

        if len(self.df) == 0:
            logger.error("Dataframe empty. Is the CSV file ok?")

        self.df = self.df.astype('object')

        self.header = list(self.df.columns.values)

    def __len__(self):

        return len(self.df.index)

    def parse_properties(self, df, idx):

        properties = {}
        if len(self.props_keys) > 0:
            for key in self.props_keys:
                # Handle Date data type
                if key.startswith("date:"):
                    prop = key[len("date:"):]  # remove prefix
                    properties[prop] = {"_date": self.df.loc[idx, key]}
                else:
                    value = self.df.loc[idx, key]
                    if value == value:  # skips nan values
                        properties[key] = value

        return properties

    def parse_constraints(self, df, idx):

        constraints = {}
        if len(self.constraints_keys) > 0:
            for key in self.constraints_keys:
                if key.startswith("constraint_date:"):
                    prop = key[len("constraint_date:"):]  # remove prefix
                    constraints[prop] = [
                        "==", {"_date": self.df.loc[idx, key]}]
                else:
                    prop = key[len(CONTRAINTS_PREFIX):]  # remove "prefix
                    constraints[prop] = ["==", self.df.loc[idx, key]]

        return constraints

    def _basic_command(self, idx, custom_fields: dict = None):
        if custom_fields == None:
            custom_fields = {}
        properties = self.parse_properties(self.df, idx)
        constraints = self.parse_constraints(self.df, idx)
        query = {
            self.command: custom_fields
        }
        if properties:
            query[self.command][PROPERTIES] = properties

        if constraints:
            query[self.command]["if_not_found"] = constraints

        return query

    def __getitem__(self, subscript):
        if isinstance(subscript, slice):
            start = subscript.start if subscript.start else 0
            stop = subscript.stop if subscript.stop else len(self)
            step = subscript.step if subscript.step else 1
            return [self.getitem(i) for i in range(start, stop, step)]
        else:
            return self.getitem(subscript)

    def getitem(self, subscript):
        Exception("getitem not implemented")

    def validate(self):

        Exception("Validation not implemented!")



import time
import requests
import boto3

import numpy as np
import cv2

from aperturedb import CSVParser
import logging

logger = logging.getLogger(__name__)

HEADER_PATH   = "filename"
HEADER_URL    = "url"
HEADER_S3_URL = "s3_url"
PROPERTIES    = "properties"
CONSTRAINTS   = "constraints"
IMG_FORMAT    = "format"


class ImageDataCSV(CSVParser.CSVParser):
    """**ApertureDB Image Data.**

    This class loads the Image Data which is present in a csv file,
    and converts it into a series of aperturedb queries.


    .. note::
        Is backed by a csv file with the following columns (format optional):

            ``filename``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

            OR

            ``url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

            OR

            ``s3_url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``
            ...

    Example csv file::

        filename,id,label,constraint_id,format
        /home/user/file1.jpg,321423532,dog,321423532,jpg
        /home/user/file2.jpg,42342522,cat,42342522,png
        ...

    Example usage:

    .. code-block:: python

        data = ImageDataCSV("/path/to/ImageData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)


    .. important::
        In the above example, the constraint_id ensures that an Image with the specified
        id would be only inserted if it does not already exist in the database.


    """

    def __init__(self, filename, check_image=True, n_download_retries=3):

        super().__init__(filename)

        self.check_image = check_image

        self.format_given     = IMG_FORMAT in self.header
        self.props_keys       = [x for x in self.header[1:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.props_keys       = [x for x in self.props_keys if x != IMG_FORMAT]
        self.constraints_keys = [x for x in self.header[1:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]

        self.source_type      = self.header[0]
        loaders = [self.load_image, self.load_url, self.load_s3_url]
        source_types = [HEADER_PATH, HEADER_URL, HEADER_S3_URL]
        if self.source_type not in source_types:
            logger.error("Source not recognized: " + self.source_type)
            raise Exception("Error loading image: " + filename)
        self.source_loader    = {
            st: sl for st, sl in zip(source_types, loaders)
        }

        self.n_download_retries = n_download_retries
        self.command = "AddImage"

    def getitem(self, idx):
        image_path = self.df.loc[idx, self.source_type]
        img_ok, img = self.source_loader[self.source_type](image_path)

        if not img_ok:
            logger.error("Error loading image: " + image_path)
            raise Exception("Error loading image: " + image_path)

        q = []
        blobs = []
        custom_fields = {}
        if self.format_given:
            custom_fields["format"] = self.df.loc[idx, IMG_FORMAT]
        ai = self._basic_command(idx, custom_fields)
        blobs.append(img)
        q.append(ai)

        return q, blobs

    def load_image(self, filename):
        if self.check_image:
            try:
                a = cv2.imread(filename)
                if a.size <= 0:
                    logger.error("IMAGE SIZE ERROR:", filename)
                    return False, None
            except Exception as e:
                logger.error("IMAGE ERROR:", filename)
                logger.exception(e)

        try:
            fd = open(filename, "rb")
            buff = fd.read()
            fd.close()
            return True, buff
        except Exception as e:
            logger.error("IMAGE ERROR:", filename)
            logger.exception(e)
        return False, None

    def check_image_buffer(self, img):
        try:
            decoded_img = cv2.imdecode(img, cv2.IMREAD_COLOR)

            # Check image is correct
            decoded_img = decoded_img if decoded_img is not None else img

            return True
        except:
            return False

    def load_url(self, url):
        retries = 0
        while True:
            imgdata = requests.get(url)
            if imgdata.ok:
                imgbuffer = np.frombuffer(imgdata.content, dtype='uint8')
                if self.check_image and not self.check_image_buffer(imgbuffer):
                    logger.error("IMAGE ERROR: ", url)
                    return False, None

                return imgdata.ok, imgdata.content
            else:
                if retries >= self.n_download_retries:
                    break
                logger.warning("WARNING: Retrying object:", url)
                retries += 1
                time.sleep(2)

        return False, None

    def load_s3_url(self, s3_url):
        retries = 0

        # The connections by boto3 cause ResourceWarning. Known
        # issue: https://github.com/boto/boto3/issues/454
        s3 = boto3.client('s3')

        while True:
            try:
                bucket_name = s3_url.split("/")[2]
                object_name = s3_url.split("s3://" + bucket_name + "/")[-1]
                s3_response_object = s3.get_object(
                    Bucket=bucket_name, Key=object_name)
                img = s3_response_object['Body'].read()
                imgbuffer = np.frombuffer(img, dtype='uint8')
                if self.check_image and not self.check_image_buffer(imgbuffer):
                    logger.error("IMAGE ERROR: ", s3_url)
                    return False, None

                return True, img
            except:
                if retries >= self.n_download_retries:
                    break
                logger.warning("WARNING: Retrying object:", s3_url)
                retries += 1
                time.sleep(2)

        logger.error("S3 ERROR:", s3_url)
        return False, None

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] not in [HEADER_PATH, HEADER_URL, HEADER_S3_URL]:
            raise Exception(
                "Error with CSV file field: filename. Must be first field")



from aperturedb import ParallelLoader
from aperturedb import CSVParser

ENTITY_CLASS = "EntityClass"
PROPERTIES   = "properties"
CONSTRAINTS  = "constraints"


class EntityGeneratorCSV(CSVParser.CSVParser):
    """**ApertureDB Entity Data loader.**

    .. warning::
        Deprecated. Use :class:`~aperturedb.EntityDataCSV.EntityDataCSV` instead.

    .. note::
        Expects a csv file with the following columns:

            ``EntityClass``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``

    Example csv file::

        EntityClass,name,lastname,age,id,constaint_id
        Person,John,Salchi,69,321423532,321423532
        Person,Johna,Salchi,63,42342522,42342522
        ...
    """

    def __init__(self, filename):

        super().__init__(filename)

        self.props_keys       = [x for x in self.header[1:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.constraints_keys = [x for x in self.header[1:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]

    def getitem(self, idx):

        data = {}
        data["class"] = self.df.loc[idx, ENTITY_CLASS]

        properties  = self.parse_properties(self.df, idx)
        constraints = self.parse_constraints(self.df, idx)

        if properties:
            data[PROPERTIES] = properties

        if constraints:
            data[CONSTRAINTS] = constraints

        return data

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] != ENTITY_CLASS:
            raise Exception("Error with CSV file field: " + ENTITY_CLASS)


class EntityLoader(ParallelLoader.ParallelLoader):
    """**ApertureDB Entity Loader.**

        This class is to be used in combination with a "generator".
        The generator must be an iterable object that generated "entity_data"
        elements.

    Example::

            entity_data = {
                "class":       entity_class,
                "properties":  properties,
                "constraints": constraints,
            }
    """

    def __init__(self, db, dry_run=False):

        super().__init__(db, dry_run=dry_run)

        self.type = "entity"

    def generate_batch(self, entity_data):

        q = []
        blobs = []

        for data in entity_data:

            ae = {
                "AddEntity": {
                    "class": data["class"]
                }
            }

            if PROPERTIES in data:
                ae["AddEntity"][PROPERTIES] = data[PROPERTIES]

            if CONSTRAINTS in data:
                ae["AddEntity"]["if_not_found"] = data[CONSTRAINTS]

            q.append(ae)

        if self.dry_run:
            print(q)

        return q, blobs



import argparse
import random
import math

from datetime import datetime

import pandas as pd
import numpy as np

from itertools import product


def generate_person_csv(multiplier):

    names     = ["James", "Luis", "Sole", "Maria",
                 "Tom", "Xavi", "Dimitris", "King"]
    lastnames = ["Ramirez", "Berlusconi", "Copola", "Tomson", "Ferro"]
    names     = names     * multiplier
    lastnames = lastnames * multiplier

    persons  = list(product(names, lastnames))

    entity   = ["Person" for x in range(len(persons))]
    ids      = [int(1000000000 * random.random()) for i in range(len(persons))]
    age      = [int(100 * random.random()) for i in range(len(persons))]
    height   = [float(200 * random.random()) for i in range(len(persons))]
    dog      = [x > 100 for x in height]
    birth    = [datetime.now().isoformat() for x in range(len(persons))]

    df = pd.DataFrame(persons, columns=['name', 'lastname'])
    df["EntityClass"] = entity
    df = df.reindex(["EntityClass", "name", "lastname"], axis=1)
    df["id"]       = ids
    df["age"]      = age
    df["height"]   = height
    df["has_dog"]  = dog
    df["date:dob"] = birth
    df["constraint_id"] = ids

    df = df.sort_values("id")

    df.to_csv("input/persons.adb.csv", index=False)

    return df


def generate_blobs_csv():

    path       = "input/blobs/"
    blob_paths = [path + str(x).zfill(4) + ".blob" for x in range(20)]
    license    = [x for x in range(10)]
    blobs      = list(product(blob_paths, license))

    # generate the blobs
    arr = [1.5, 2.4, 3.3, 5.5, 9.9, 111.12, 1000.20]
    nparr = np.array(arr)
    for blob in blob_paths:
        blobfile = open(blob, 'wb')
        nparr.tofile(blobfile)
        blobfile.close()

    entity   = ["segmentation" for x in range(len(blobs))]
    ids      = [int(1000000000 * random.random()) for i in range(len(blobs))]

    df = pd.DataFrame(blobs, columns=['filename', 'license'])
    df                  = df.reindex(["filename", "license"], axis=1)
    df["id"]            = ids
    df["constraint_id"] = ids

    df = df.sort_values("id")

    df.to_csv("input/blobs.adb.csv", index=False)

    return df


def generate_images_csv(multiplier):

    multiplier = multiplier // 2
    path    = "input/images/"
    imgs    = [path + str(x).zfill(4) +
               ".jpg" for x in range(200)] * multiplier
    license = [x for x in range(10)] * multiplier

    images  = list(product(imgs, license))

    ids      = [int(1000000000 * random.random()) for i in range(len(images))]
    age      = [int(100 * random.random()) for i in range(len(images))]
    height   = [float(200 * random.random()) for i in range(len(images))]
    dog      = [x > 100 for x in height]
    date_cap = [datetime.now().isoformat() for x in range(len(images))]

    df = pd.DataFrame(images, columns=['filename', 'license'])
    df["id"]       = ids
    df["age"]      = age
    df["height"]   = height
    df["has_dog"]  = dog
    df["date:date_captured"] = date_cap
    df["constraint_id"] = ids

    df = df.sort_values("id")

    df.to_csv("input/images.adb.csv", index=False)

    return df


def generate_http_images_csv(ip_file_csv):

    images    = pd.read_csv(ip_file_csv, sep=",", header=None)

    ids      = [int(1000000000 * random.random()) for i in range(len(images))]
    age      = [int(100 * random.random()) for i in range(len(images))]
    height   = [float(200 * random.random()) for i in range(len(images))]
    license  = [x for x in range(len(images))]

    df = pd.DataFrame()
    df['url']      = images
    df["urlid"]    = ids
    df['license']  = license
    df["age"]      = age
    df["height"]   = height
    df["constraint_urlid"] = ids

    df = df.sort_values("urlid")

    df.to_csv("input/http_images.adb.csv", index=False)

    return df


def generate_s3_images_csv(ip_file_csv):

    images    = pd.read_csv(ip_file_csv, sep=",", header=None)

    ids      = [int(1000000000 * random.random()) for i in range(len(images))]
    age      = [int(100 * random.random()) for i in range(len(images))]
    height   = [float(200 * random.random()) for i in range(len(images))]
    license  = [x for x in range(len(images))]

    df = pd.DataFrame()
    df['s3_url']   = images
    df["id"]       = ids
    df['license']  = license
    df["age"]      = age
    df["height"]   = height
    df["constraint_id"] = ids

    df = df.sort_values("id")

    df.to_csv("input/s3_images.adb.csv", index=False)

    return df


def generate_connections_csv(persons, images):

    connections  = list(product(images["id"][::100], persons["id"][::100]))

    connect   = ["has_image" for x in range(len(connections))]
    confidence  = [random.random() for x in range(len(connections))]

    df = pd.DataFrame(connections, columns=['_Image@id', 'Person@id'])
    df["ConnectionClass"] = connect
    df = df.reindex(["ConnectionClass", '_Image@id', 'Person@id'], axis=1)

    df["confidence"]      = confidence

    df.to_csv("input/connections-persons-images.adb.csv", index=False)


def generate_bboxes_csv(images):

    images  = images["id"]

    x   = [0 for x in range(len(images))]
    y   = [0 for x in range(len(images))]
    w   = [100 for x in range(len(images))]
    h   = [150 for x in range(len(images))]

    labels = ["dog", "cat", "catdog", "rocko", "philip", "froghead", "drog"]

    confidence  = [random.random() for x in range(len(images))]
    label       = [labels[math.floor(random.random() * len(labels))]
                   for x in range(len(images))]

    df = pd.DataFrame(images, columns=['id'])

    df["x_pos"] = x
    df["y_pos"] = y
    df["width"] = w
    df["height"] = h
    df["confidence"] = confidence
    df["label"]      = label

    df = df.sort_values("id")
    df.to_csv("input/bboxes.adb.csv", index=False)


def generate_bboxes_constraints_csv(images):
    image_id = images.iloc[0]["id"]
    num_boxes = 3

    x   = [0 for x in range(num_boxes)]
    y   = [0 for x in range(num_boxes)]
    w   = [100 for x in range(num_boxes)]
    h   = [150 for x in range(num_boxes)]

    labels = ["dog", "cat", "catdog", "rocko", "philip", "froghead", "drog"]

    confidence  = [random.random() for x in range(num_boxes)]
    label       = [labels[math.floor(random.random() * num_boxes)]
                   for x in range(num_boxes)]

    df = pd.DataFrame([{
        "id": image_id,
    } for _ in range(num_boxes)])

    df["x_pos"] = x
    df["y_pos"] = y
    df["width"] = w
    df["height"] = h
    df["confidence"] = confidence
    df["label"]      = label
    df["box_id"] = [123] * num_boxes
    df["constraint_box_id"] = [123] * num_boxes

    df = df.sort_values("id")
    df.to_csv("input/bboxes-constraints.adb.csv", index=False)


def generate_descriptors(images, setname, dims):

    filename = "input/" + setname + "_desc.npy"

    images  = images["id"]

    descriptors = np.random.rand(len(images), dims)

    np.save(filename, descriptors)

    labels = ["dog", "cat", "catdog", "rocko", "philip", "froghead", "drog"]
    confidence  = [random.random() for x in range(len(images))]
    label       = [labels[math.floor(random.random() * len(labels))]
                   for x in range(len(images))]

    df = pd.DataFrame(images, columns=['id'])

    df["filename"] = [filename for x in range(len(images))]
    df["index"]    = [i for i in range(len(images))]
    df["set"]      = [setname for x in range(len(images))]
    df["id"]       = [i for i in range(len(images))]
    df["label"]      = label
    df["confidence"] = confidence

    df = df.sort_values("id")
    df = df.reindex(columns=["filename", "index",
                    "set", "id", "label", "confidence"])
    df.to_csv("input/" + setname + ".adb.csv", index=False)


def generate_descriptorset(names, dims):

    metrics = ["L2", "IP", "CS", ["CS", "IP"], ["L2", "CS", "IP"]]
    engines = ["FaissIVFFlat", "FaissFlat", ["FaissIVFFlat", "FaissFlat"]]

    df = pd.DataFrame()

    df["name"]       = names
    df["dimensions"] = dims
    df["engine"]     = [random.sample(engines, 1)[0] for i in names]
    df["metric"]     = [random.sample(metrics, 1)[0] for i in names]

    df.to_csv("input/descriptorset.adb.csv", index=False)


def main(params):

    persons = generate_person_csv(params.multiplier)
    blobs   = generate_blobs_csv()
    images  = generate_images_csv(int(params.multiplier / 2))
    s3_imgs = generate_http_images_csv("input/sample_http_urls.csv")
    s3_imgs = generate_s3_images_csv("input/sample_s3_urls.csv")
    connect = generate_connections_csv(persons, images)
    bboxes  = generate_bboxes_csv(images)
    bboxes_constraints = generate_bboxes_constraints_csv(images)

    desc_name = ["setA", "setB", "setC", "setD", "setE", "setF"]
    desc_dims = [2048, 1025, 2048, 1025, 2048, 1025]    # yes, 1025
    generate_descriptorset(desc_name, desc_dims)
    for name, dims in zip(desc_name, desc_dims):
        generate_descriptors(images, name, dims)


def get_args():
    obj = argparse.ArgumentParser()

    # Run Config
    obj.add_argument('-multiplier', type=int, default=10)

    params = obj.parse_args()

    return params


if __name__ == "__main__":
    args = get_args()
    main(args)



import cv2
import logging
from aperturedb import CSVParser

logger = logging.getLogger(__name__)

HEADER_PATH = "filename"
PROPERTIES  = "properties"
CONSTRAINTS = "constraints"


class VideoDataCSV(CSVParser.CSVParser):
    """
    **ApertureDB Video Data.**

    This class loads the Video Data which is present in a csv file,
    and converts it into a series of aperturedb queries.


    .. note::

        Expects a csv file with the following columns:

            ``filename``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``

    Example csv file::

        filename,id,label,constaint_id
        /home/user/file1.jpg,321423532,dog,321423532
        /home/user/file2.jpg,42342522,cat,4234252
        ...

    Example usage:

    .. code-block:: python

        data = ImageDataCSV("/path/to/VideoData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)
    """

    def __init__(self, filename, check_video=True):

        super().__init__(filename)

        self.check_video = check_video

        self.props_keys       = [x for x in self.header[1:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.constraints_keys = [x for x in self.header[1:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.command = "AddVideo"

    def getitem(self, idx):
        filename   = self.df.loc[idx, HEADER_PATH]
        video_ok, video = self.load_video(filename)

        if not video_ok:
            logger.error("Error loading video: " + filename)
            raise Exception("Error loading video: " + filename)

        q = []
        blobs = []
        av = self._basic_command(idx)
        blobs.append(video)
        q.append(av)

        return q, blobs

    def load_video(self, filename):

        if self.check_video:
            try:
                a = cv2.VideoCapture(filename)
                if a.isOpened() == False:
                    logger.error("Video reading Error:", filename)
            except Exception as e:
                logger.error("Video Error:", filename)
                logger.exception(e)

        try:
            fd = open(filename, "rb")
            buff = fd.read()
            fd.close()
            return True, buff
        except Exception as e:
            logger.error("Video Error:", filename)
            logger.exception(e)

        return False, None

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] != HEADER_PATH:
            raise Exception(
                "Error with CSV file field: filename. Must be first field")



import argparse
import random
import math

from aperturedb import ImageDownloader


def main(params):

    downloader = ImageDownloader.ImageDownloader(
        check_if_present=True, n_download_retries=2)
    downloader.run(ImageDownloader.ImageDownloaderCSV(params.in_file),
                   numthreads=32,
                   batchsize=1,
                   stats=True)


def get_args():
    obj = argparse.ArgumentParser()

    # Run Config
    obj.add_argument('-in_file', type=str, default="input/url_images.adb.csv")

    params = obj.parse_args()

    return params


if __name__ == "__main__":
    args = get_args()
    main(args)



from aperturedb import ParallelLoader
from aperturedb import CSVParser

CONNECTION_CLASS = "ConnectionClass"
PROPERTIES  = "properties"
CONSTRAINTS = "constraints"


class ConnectionGeneratorCSV(CSVParser.CSVParser):
    """**ApertureDB Connection Data generator.**

    .. warning::
        Deprecated. Use :class:`~aperturedb.ConnectionDataCSV.ConnectionDataCSV` instead.

    .. note::
        Is backed by a csv file with the following columns:

            ``ConnectionClass``, ``Class1@PROP_NAME``, ``Class2@PROP_NAME`` ... ``PROP_NAME_N``, ``constraint_PROP1``

    Example csv file::

        ConnectionClass,_Image@id,_Descriptor@UUID,confidence,id,constaint_id
        has_image,321423532,AID-0X3E,0.4354,5432543254,5432543254
        has_image,42342522,BXY-AB1Z,0.6432,98476542,98476542
        ...

    **ConnectionClass**: Arbitrary class name for the entity this would be saved as.

    **<ClassName>@<PropertyName>** This is a special combination of Class Name and
    Property Name that can uniquely identify an entity.
    '@' is a delimeter, so should not be used in a property name.

    **PROP_NAME_1 .. PROP_NAME_N** Arbitraty property names.

    **constraint_PROP_NAME_1** A equality check against a unique property to ensure duplicates are not inserted.
    """

    def __init__(self, filename):

        super().__init__(filename)

        self.props_keys       = [x for x in self.header[3:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]

        self.constraints_keys = [x for x in self.header[3:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]

        self.src_class   = self.header[1].split("@")[0]
        self.src_key     = self.header[1].split("@")[1]
        self.dst_class   = self.header[2].split("@")[0]
        self.dst_key     = self.header[2].split("@")[1]

    def getitem(self, idx):

        data = {
            "class":      self.df.loc[idx, CONNECTION_CLASS],
            "src_class": self.src_class,
            "src_key":   self.src_key,
            "src_val":   self.df.loc[idx, self.header[1]],
            "dst_class": self.dst_class,
            "dst_key":   self.dst_key,
            "dst_val":   self.df.loc[idx, self.header[2]],
        }

        properties  = self.parse_properties(self.df, idx)
        constraints = self.parse_constraints(self.df, idx)

        if properties:
            data[PROPERTIES] = properties

        if constraints:
            data[CONSTRAINTS] = constraints

        return data

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] != CONNECTION_CLASS:
            raise Exception(
                "Error with CSV file field: 0 - " + CONNECTION_CLASS)
        if "@" not in self.header[1]:
            raise Exception("Error with CSV file field: 1")
        if "@" not in self.header[2]:
            raise Exception("Error with CSV file field: 2")


class ConnectionLoader(ParallelLoader.ParallelLoader):
    """**ApertureDB Connection Loader.**

    This executes in conjunction with a **generator** object,
    for example :class:`~aperturedb.ConnectionLoader.ConnectionGeneratorCSV`,
    which is a class that implements iterable inteface and generates "connection data" elements.

    Example::

            connection_data = {
                "class":       connection_class,
                "src_class": "_Image",
                "src_key":   "id",
                "src_val":   "321423532",
                "dst_class": "_Descriptor",
                "dst_key":   "UUID",
                "dst_val":   "AID-0X3E",
                "properties":  properties,
                "constraints": constraints,
            }
    """

    def __init__(self, db, dry_run=False):

        super().__init__(db, dry_run=dry_run)

        self.type = "connection"

    def generate_batch(self, entity_data):

        q = []

        ref_counter = 1

        for data in entity_data:

            try:

                ref_src = ref_counter
                ref_counter += 1
                fe_a = {
                    "FindEntity": {
                        "_ref": ref_src,
                        "with_class": data["src_class"],
                    }
                }

                fe_a["FindEntity"]["constraints"] = {}
                fe_a["FindEntity"]["constraints"][data["src_key"]] = [
                    "==", data["src_val"]]
                q.append(fe_a)

                ref_dst = ref_counter
                ref_counter += 1
                fe_b = {
                    "FindEntity": {
                        "_ref": ref_dst,
                        "with_class": data["dst_class"],
                    }
                }

                fe_b["FindEntity"]["constraints"] = {}
                fe_b["FindEntity"]["constraints"][data["dst_key"]] = [
                    "==", data["dst_val"]]
                q.append(fe_b)

                ae = {
                    "AddConnection": {
                        "class": data["class"],
                        "src": ref_src,
                        "dst": ref_dst,
                    }
                }

                if PROPERTIES in data:
                    ae["AddConnection"][PROPERTIES] = data[PROPERTIES]

                if CONSTRAINTS in data:
                    ae["AddConnection"]["if_not_found"] = data[CONSTRAINTS]

                q.append(ae)
            except KeyError as ex:
                error_str = "ERROR: ConnectionLoader::generate_batch():" + \
                    " expected key='" + ex.args[0] + "' in '" + str(data) + \
                    "'. Ignored."
                print(error_str)
                sys.stdout.write(error_str + "\n")

        if self.dry_run:
            print(q)

        return q, []



import ast
from aperturedb import ParallelLoader
from aperturedb import CSVParser

HEADER_NAME   = "name"
HEADER_DIM    = "dimensions"
HEADER_ENGINE = "engine"
HEADER_METRIC = "metric"
PROPERTIES    = "properties"
CONSTRAINTS   = "constraints"


class DescriptorSetDataCSV(CSVParser.CSVParser):
    """**ApertureDB DescriptorSet Data.**

    This class loads the Descriptor Set Data which is present in a csv file,
    and converts it into a series of aperturedb queries.

    .. note::
        Is backed by a csv file with the following columns:

            ``name``, ``dimensions``, ``engine``, ``metric``, ``PROP_NAME_N``, ``constraint_PROP1``

    Example csv file::

        name,dimensions,engine,metric
        dining_chairs,2048,FaissIVFFlat,L2
        chandeliers,2048,FaissIVFFlat,L2
        console_tables,2048,FaissIVFFlat,L2
        ...

    Example code to create an instance:

    .. code-block:: python

        data = DescriptorSetDataCSV("/path/to/DescriptorSetData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)


    .. important::
        In the above example, the first row implies to create a Descriptor set called dining_chairs.
        The Descriptors in that set wouldb be expected to be an array of float64, of length 2048.
        When performing a search on this set, FaissIVFFlat engine would be used and the metric to compute
        the distance would be L2.
    """

    def __init__(self, filename):

        super().__init__(filename)

        self.props_keys       = [x for x in self.header[4:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.constraints_keys = [x for x in self.header[4:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.command = "AddDescriptorSet"

    def getitem(self, idx):

        # Metrics/Engine can be of the form:
        #       "IP", or
        #       ["IP" ...]

        metrics = self.df.loc[idx, HEADER_METRIC]
        metrics = metrics if "[" not in metrics else ast.literal_eval(metrics)
        engines = self.df.loc[idx, HEADER_ENGINE]
        engines = engines if "[" not in engines else ast.literal_eval(engines)

        data = {
            "name":       self.df.loc[idx, HEADER_NAME],
            "dimensions": self.df.loc[idx, HEADER_DIM],
            "engine":     engines,
            "metric":     metrics,
        }

        q = []
        ads = self._basic_command(idx, custom_fields=data)
        q.append(ads)

        return q, []

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] != HEADER_NAME:
            raise Exception("Error with CSV file field: " + HEADER_NAME)
        if self.header[1] != HEADER_DIM:
            raise Exception("Error with CSV file field: " + HEADER_DIM)
        if self.header[2] != HEADER_ENGINE:
            raise Exception("Error with CSV file field: " + HEADER_ENGINE)
        if self.header[3] != HEADER_METRIC:
            raise Exception("Error with CSV file field: " + HEADER_METRIC)



from aperturedb import CSVParser

ENTITY_CLASS = "EntityClass"
PROPERTIES   = "properties"
CONSTRAINTS  = "constraints"


class EntityDataCSV(CSVParser.CSVParser):
    """**ApertureDB Entity Data.**

    This class loads the Entity Data which is present in a csv file,
    and converts it into a series of aperturedb queries.

    .. note::
        Expects a csv file with the following columns:

            ``EntityClass``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``

    Example csv file::

        EntityClass,name,lastname,age,id,constraint_id
        Person,John,Salchi,69,321423532,321423532
        Person,Johna,Salchi,63,42342522,42342522
        ...

    Example usage:

    .. code-block:: python

        data = EntityDataCSV("/path/to/EntityData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)


    .. important::
        In the above example, the constraint_id ensures that a Entity with the specified
        id would be only inserted if it does not already exist in the database.


    """

    def __init__(self, filename):

        super().__init__(filename)

        self.props_keys       = [x for x in self.header[1:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.constraints_keys = [x for x in self.header[1:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.command = "AddEntity"

    def getitem(self, idx):
        eclass = self.df.loc[idx, ENTITY_CLASS]
        q = []
        ae = self._basic_command(idx,
                                 custom_fields = {
                                     "class": eclass
                                 })

        q.append(ae)
        return q, []

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] != ENTITY_CLASS:
            raise Exception("Error with CSV file field: " + ENTITY_CLASS)



from typing import List, Tuple
from torch.utils.data import Dataset


class PytorchData(object):
    def __init__(self, dataset: Dataset) -> None:
        self.loaded_dataset = [t for t in dataset]

    def __getitem__(self, idx: int):
        return self.generate_query(idx)

    def __len__(self):
        return len(self.loaded_dataset)

    def generate_query(self, idx: int) -> Tuple[List[dict], List[bytes]]:
        """
        **Takes information from one atomic record from the Data and converts it to Query for apertureDB**

        Args:
            idx (int): index of the record in collection.

        Raises:
            Exception: _description_

        Returns:
            Tuple[List[dict], List[bytes]]: A pair of list of commands and optional list of blobs to go with them.
        """
        raise Exception("To be implemented by subclass")



import time

from test_Base import TestBase

import logging
logger = logging.getLogger(__name__)


class TestSession(TestBase):

    '''
        These check operation of the Connector Session
    '''

    def test_sessionRenew(self):
        '''
            Verifies that Session renew works
        '''

        try:
            db = self.create_connection()
            # force session token expiry
            db.session.session_token_ttl = 1
            logging.debug("Connected? {0}".format(
                "yes" if db.connected else "no"))
            logging.debug(
                "Session valid? {0}".format(
                    "yes" if db.session.valid() else "no"))
            logging.debug("Valid length: {0}".format(
                db.session.session_token_ttl))
            time.sleep(2)
            query = [{
                "FindImage": {
                    "results": {
                        "limit": 5
                    }
                }
            }]
            responses, blobs = db.query(query)
            logging.debug(responses)
            self.assertTrue(db.session.valid(), "Failed to renew Session")
        except Exception:
            self.fail("Failed to renew Session")




from typing import List, Tuple
import os
import pandas as pd
from kaggle.api.kaggle_api_extended import KaggleApi
import zipfile


class KaggleData(object):
    """
    **Class to wrap around a Dataset retrieved from kaggle**

    A DataSet downloaded from kaggle does not implement a standard mechanism to iterate over it's values
    This class intends to provide an abstracion like that of a pytorch dataset.
    Where the iteration over Dataset elements yields an atomic record.

    .. note::
        This class should be subclassed with specefic implementations of generate_index and generate_query.

    Example subclass:

    .. code-block:: python

        class CelebADataKaggle(KaggleData):
           def __init__(self, **kwargs) -> None:
               self.records_count = kwargs["records_count"]
               self.embedding_generator = kwargs["embedding_generator"]
               self.search_set_name = kwargs["search_set_name"]
               self.records_count = kwargs["records_count"]
               super().__init__(dataset_ref = "jessicali9530/celeba-dataset",
                                records_count=self.records_count)

           def generate_index(self, root: str, records_count=-1) -> pd.DataFrame:
               attr_index = pd.read_csv(
                   os.path.join(root, "list_attr_celeba.csv"))
               bbox_index = pd.read_csv(
                   os.path.join(root, "list_bbox_celeba.csv"))
               landmarks_index = pd.read_csv(os.path.join(
                   root, "list_landmarks_align_celeba.csv"))
               partition_index = pd.read_csv(
                   os.path.join(root, "list_eval_partition.csv"))
               rows = attr_index.combine_first(bbox_index).combine_first(
                   landmarks_index).combine_first(partition_index)
               original_size = len(rows)
               records_count = records_count if records_count > 0 else original_size

               rows = rows[:records_count]

               print(
                   f"Created {len(rows)} items from {original_size} in the original dataset.")
               return rows

           def generate_query(self, idx: int) -> Tuple[List[dict], List[bytes]]:
               record = self.collection[idx]
               p = record
               q = [
                   {
                       "AddImage": {
                           "_ref": 1,
                           "properties": {
                               c: p[c] for c in p.keys()
                           },
                       }
                   }, {
                       "AddBoundingBox": {
                           "_ref": 2,
                           "image": 1,
                           "rectangle": {
                               "x": p["x_1"],
                               "y": p["y_1"],
                               "width": p["width"],
                               "height": p["height"]
                           }
                       }
                   }, {
                       "AddDescriptor": {
                           "set": self.search_set_name,
                           "connect": {
                               "ref": 1
                           }
                       }
                   }
               ]
               q[0]["AddImage"]["properties"]["keypoints"] = f"10 {p['lefteye_x']} {p['lefteye_y']} {p['righteye_x']} {p['righteye_y']} {p['nose_x']} {p['nose_y']} {p['leftmouth_x']} {p['leftmouth_y']} {p['rightmouth_x']} {p['rightmouth_y']}"
               image_file_name = os.path.join(
                   self.workdir,
                   'img_align_celeba/img_align_celeba',
                   p["image_id"])
               blob = open(image_file_name, "rb").read()
               embedding = self.embedding_generator(Image.open(image_file_name))
               serialized = embedding.cpu().detach().numpy().tobytes()
               return q, [blob, serialized]

        Args:
            dataset_ref (str): URL of kaggle dataset, for example 'https://www.kaggle.com/datasets/crawford/cat-dataset'
            records_count (int): number of records to provide to generate.

    """

    def __init__(
            self,
            dataset_ref: str,
            records_count: int = -1) -> None:
        self._collection = None
        self.records_count = records_count
        kaggle = KaggleApi()
        kaggle.authenticate()
        if "datasets/" in dataset_ref:
            dataset_ref = dataset_ref[dataset_ref.index(
                "datasets/") + len("datasets/"):]

        workdir = os.path.join("kaggleds", dataset_ref)

        files = kaggle.dataset_list_files(dataset_ref)

        # do not unzip from kaggle's API as it deletes the archive and
        # a subsequent run results in a redownload.
        x = kaggle.dataset_download_files(
            dataset=dataset_ref,
            path=workdir,
            unzip=False)

        archive = None
        for _, subdirs, dfiles in os.walk(workdir):
            if len(dfiles) == 1 and len(subdirs) == 0:
                archive = os.path.join(workdir, dfiles[0])

                with zipfile.ZipFile(archive, 'r') as zip_ref:
                    zip_ref.extractall(workdir)

                break
        self.workdir = workdir
        self.collection = self.generate_index(
            workdir, self.records_count).to_dict('records')

    def __getitem__(self, subscript):
        if isinstance(subscript, slice):
            start = subscript.start if subscript.start else 0
            stop = subscript.stop if subscript.stop else len(self)
            step = subscript.step if subscript.step else 1
            return [self.generate_query(i) for i in range(start, stop, step)]
        if subscript >= len(self.collection):
            raise StopIteration
        return self.generate_query(subscript)

    def __len__(self):
        return len(self.collection)

    def generate_index(self, root: str, records_count: int = -1) -> pd.DataFrame:
        """**Generate a way to access each record downloaded at the root**

        Args:
            root (str): Path to wich kaggle downloads a Dataset.

        Returns:
            pd.DataFrame: The Data loaded in a dataframe.
        """
        raise Exception("To be implemented by subclass")

    def generate_query(self, idx: int) -> Tuple[List[dict], List[bytes]]:
        """
        **Takes information from one atomic record from the Data and converts it to Query for apertureDB**

        Args:
            idx (int): index of the record in collection.

        Raises:
            Exception: _description_

        Returns:
            Tuple[List[dict], List[bytes]]: A pair of list of commands and optional list of blobs to go with them.
        """
        raise Exception("To be implemented by subclass")



from aperturedb import ParallelLoader
from aperturedb import CSVParser

HEADER_X_POS  = "x_pos"
HEADER_Y_POS  = "y_pos"
HEADER_WIDTH  = "width"
HEADER_HEIGHT = "height"
IMG_KEY_PROP  = "img_key_prop"
IMG_KEY_VAL   = "img_key_value"


class BBoxGeneratorCSV(CSVParser.CSVParser):
    """**ApertureDB Bounding Box Data generator.**

    .. warning::
        Deprecated. Use :class:`~aperturedb.BBoxDataCSV.BBoxDataCSV` instead.

    .. note::
        Is backed by a csv file with the following columns:

            ``IMG_KEY``, ``x_pos``, ``y_pos``, ``width``, ``height``, ``BBOX_PROP_NAME_1``, ... ``BBOX_PROP_NAME_N``, ``constraint_BBOX_PROP_NAME_1``

    **IMG_KEY**: column has the property name of the image property that
    the bounding box will be connected to, and each row has the value
    that will be used for finding the image.

    **x_pos, y_pos**: Specify the coordinates of top left of the bounding box.

    **width, height**: Specify the dimensions of the bounding box, as integers (unit is in pixels).

    **BBOX_PROP_NAME_N**: is an arbitrary name of the property of the bounding
    box, and each row has the value for that property.

    **constraint_BBOX_PROP_NAME_1**: Constraints against specific property, used for conditionally adding a Bounding Box.

    Example csv file::

        img_unique_id,x_pos,y_pos,width,height,type,dataset_id,constraint_dataset_id
        d5b25253-9c1e,257,154,84,125,manual,12345,12345
        d5b25253-9c1e,7,537,522,282,manual,12346,12346
        ...
    """

    def __init__(self, filename):

        super().__init__(filename)

        self.props_keys       = [x for x in self.header[5:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.constraints_keys = [x for x in self.header[5:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]

        self.img_key = self.header[0]

    def getitem(self, idx):

        data = {
            "x":      int(self.df.loc[idx, HEADER_X_POS]),
            "y":      int(self.df.loc[idx, HEADER_Y_POS]),
            "width":  int(self.df.loc[idx, HEADER_WIDTH]),
            "height": int(self.df.loc[idx, HEADER_HEIGHT]),
        }

        val = self.df.loc[idx, self.img_key]

        if val != "":
            data[IMG_KEY_PROP] = self.img_key
            data[IMG_KEY_VAL]  = val

        properties  = self.parse_properties(self.df, idx)
        constraints = self.parse_constraints(self.df, idx)

        if properties:
            data[CSVParser.PROPERTIES] = properties

        if constraints:
            data[CSVParser.CONSTRAINTS] = constraints

        return data

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[1] != HEADER_X_POS:
            raise Exception("Error with CSV file field: " + HEADER_X_POS)
        if self.header[2] != HEADER_Y_POS:
            raise Exception("Error with CSV file field: " + HEADER_Y_POS)
        if self.header[3] != HEADER_WIDTH:
            raise Exception("Error with CSV file field: " + HEADER_WIDTH)
        if self.header[4] != HEADER_HEIGHT:
            raise Exception("Error with CSV file field: " + HEADER_HEIGHT)


class BBoxLoader(ParallelLoader.ParallelLoader):
    """
    **A loader that helps to ingest Bounding box information against existing images**

    This class facilitates the insertions of Bounding box information connecting them with
    the images already inserted in the database.

    This executes in conjunction with a **generator** object for example :class:`~aperturedb.BBoxLoader.BBoxGeneratorCSV`,
    which is a class that implements iterable inteface and generates "bounding box" elements.

    Example::

        bbox_data = {
            "x": ,
            "y": ,
            "width": ,
            "height":,
            "properties: {
                "BBOX_PROP_NAME_1": "BBOX_PROP_VALUE_1",
                .
                .
                .
                "BBOX_PROP_NAME_N": "BBOX_PROP_VALUE_N"
            }
        }

    """

    def __init__(self, db, dry_run=False):

        super().__init__(db, dry_run=dry_run)

        self.type = "bbox"

    def generate_batch(self, bbox_data):

        q = []

        ref_counter = 1
        for data in bbox_data:

            # TODO we could reuse image references within the batch
            # instead of creating a new find for every image.
            img_ref = ref_counter
            ref_counter += 1
            fi = {
                "FindImage": {
                    "_ref": img_ref,
                }
            }

            if IMG_KEY_PROP in data:
                key = data[IMG_KEY_PROP]
                val = data[IMG_KEY_VAL]
                constraints = {}
                constraints[key] = ["==", val]
                fi["FindImage"]["constraints"] = constraints

            q.append(fi)

            ai = {
                "AddBoundingBox": {
                    "image": img_ref,
                    "rectangle": {
                        "x":      data["x"],
                        "y":      data["y"],
                        "width":  data["width"],
                        "height": data["height"],
                    },
                }
            }

            if "properties" in data:
                props = data[CSVParser.PROPERTIES]
                if "_label" in props:
                    ai["AddBoundingBox"]["label"] = props["_label"]
                    props.pop("_label")

                # Check if props is not empty after removing "_label"
                if props:
                    ai["AddBoundingBox"]["properties"] = props

            q.append(ai)

        if self.dry_run:
            print(q)

        return q, []



import logging
from aperturedb import CSVParser

PROPERTIES  = "properties"
CONSTRAINTS = "constraints"
BLOB_PATH   = "filename"

logger = logging.getLogger(__name__)


class BlobDataCSV(CSVParser.CSVParser):
    """**ApertureDB Blob Data.**

    This class loads the Blob Data which is present in a csv file,
    and converts it into a series of aperturedb queries.

    .. note::
        Is backed by a csv file with the following columns:

            ``FILENAME``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP_NAME_1``

    **FILENAME**: The path of the blob object on the file system.

    **PROP_NAME_1 ... PROP_NAME_N**: Arbitrary property names associated with this blob.

    **constraint_PROP_NAME_1**: Constraints against specific property, used for conditionally adding a Blob.

    Example csv file::

        filename,name,lastname,age,id,constraint_id
        /mnt/blob1,John,Salchi,69,321423532,321423532
        /mnt/blob2,Johna,Salchi,63,42342522,42342522
        ...

    Example usage:

    .. code-block:: python

        data = BlobDataCSV("/path/to/BlobData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)



    .. important::
        In the above example, the constraint_id ensures that a blob with the specified
        id would be only inserted if it does not already exist in the database.
    """

    def __init__(self, filename):

        super().__init__(filename)

        self.props_keys       = [x for x in self.header[1:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX) and x != BLOB_PATH]
        self.constraints_keys = [x for x in self.header[1:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.command = "AddBlob"

    def getitem(self, idx):
        filename = self.df.loc[idx, BLOB_PATH]

        blob_ok, blob = self.load_blob(filename)
        if not blob_ok:
            logger.error("Error loading blob: " + filename)
            raise Exception("Error loading blob: " + filename)

        q = []
        ab = self._basic_command(idx)
        q.append(ab)

        return q, [blob]

    def load_blob(self, filename):

        try:
            fd = open(filename, "rb")
            buff = fd.read()
            fd.close()
            return True, buff
        except Exception as e:
            logger.exception(e)

        return False, None

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] != BLOB_PATH:
            raise Exception("Error with CSV file field: " + BLOB_PATH)



import time
import requests
import boto3

import numpy as np
import cv2

from aperturedb import ParallelLoader
from aperturedb import CSVParser

HEADER_PATH   = "filename"
HEADER_URL    = "url"
HEADER_S3_URL = "s3_url"
PROPERTIES    = "properties"
CONSTRAINTS   = "constraints"
IMG_FORMAT    = "format"


class ImageGeneratorCSV(CSVParser.CSVParser):
    """**ApertureDB Image Data generator.**

    .. warning::
        Deprecated. Use :class:`~aperturedb.ImageDataCSV.ImageDataCSV` instead.

    .. note::
        Is backed by a csv file with the following columns (format optional):

            ``filename``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

            OR

            ``url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

            OR

            ``s3_url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``
            ...

    Example csv file::

        filename,id,label,constaint_id,format
        /home/user/file1.jpg,321423532,dog,321423532,jpg
        /home/user/file2.jpg,42342522,cat,42342522,png
        ...


    """

    def __init__(self, filename, check_image=True, n_download_retries=3):

        super().__init__(filename)

        self.check_image = check_image

        self.format_given     = IMG_FORMAT in self.header
        self.props_keys       = [x for x in self.header[1:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.props_keys       = [x for x in self.props_keys if x != IMG_FORMAT]
        self.constraints_keys = [x for x in self.header[1:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]

        self.source_type      = self.header[0]
        if self.source_type not in [HEADER_PATH, HEADER_URL, HEADER_S3_URL]:
            print("Source not recognized: " + self.source_type)
            raise Exception("Error loading image: " + filename)

        self.n_download_retries = n_download_retries

    # TODO: we can add support for slicing here.
    def getitem(self, idx):

        data = {}

        img_ok = True
        img = None

        if self.source_type == HEADER_PATH:
            image_path   = self.df.loc[idx, HEADER_PATH]
            img_ok, img  = self.load_image(image_path)
        elif self.source_type == HEADER_URL:
            image_path   = self.df.loc[idx, HEADER_URL]
            img_ok, img  = self.load_url(image_path)
        elif self.source_type == HEADER_S3_URL:
            image_path   = self.df.loc[idx, HEADER_S3_URL]
            img_ok, img  = self.load_s3_url(image_path)

        if not img_ok:
            print("Error loading image: " + filename)
            raise Exception("Error loading image: " + filename)

        data["img_blob"] = img
        if self.format_given:
            data["format"] = self.df.loc[idx, IMG_FORMAT]

        properties  = self.parse_properties(self.df, idx)
        constraints = self.parse_constraints(self.df, idx)

        if properties:
            data[PROPERTIES] = properties

        if constraints:
            data[CONSTRAINTS] = constraints

        return data

    def load_image(self, filename):
        if self.check_image:
            try:
                a = cv2.imread(filename)
                if a.size <= 0:
                    print("IMAGE SIZE ERROR:", filename)
                    return false, None
            except:
                print("IMAGE ERROR:", filename)

        try:
            fd = open(filename, "rb")
            buff = fd.read()
            fd.close()
            return True, buff
        except:
            print("IMAGE ERROR:", filename)
        return False, None

    def check_image_buffer(self, img):
        try:
            decoded_img = cv2.imdecode(img, cv2.IMREAD_COLOR)

            # Check image is correct
            decoded_img = decoded_img if decoded_img is not None else img

            return True
        except:
            return False

    def load_url(self, url):
        retries = 0
        while True:
            imgdata = requests.get(url)
            if imgdata.ok:
                imgbuffer = np.frombuffer(imgdata.content, dtype='uint8')
                if self.check_image and not self.check_image_buffer(imgbuffer):
                    print("IMAGE ERROR: ", url)
                    return False, None

                return imgdata.ok, imgdata.content
            else:
                if retries >= self.n_download_retries:
                    break
                print("WARNING: Retrying object:", url)
                retries += 1
                time.sleep(2)

        return False, None

    def load_s3_url(self, s3_url):
        retries = 0

        # The connections by boto3 cause ResourceWarning. Known
        # issue: https://github.com/boto/boto3/issues/454
        s3 = boto3.client('s3')

        while True:
            try:
                bucket_name = s3_url.split("/")[2]
                object_name = s3_url.split("s3://" + bucket_name + "/")[-1]
                s3_response_object = s3.get_object(
                    Bucket=bucket_name, Key=object_name)
                img = s3_response_object['Body'].read()
                imgbuffer = np.frombuffer(img, dtype='uint8')
                if self.check_image and not self.check_image_buffer(imgbuffer):
                    print("IMAGE ERROR: ", s3_url)
                    return False, None

                return True, img
            except:
                if retries >= self.n_download_retries:
                    break
                print("WARNING: Retrying object:", s3_url)
                retries += 1
                time.sleep(2)

        print("S3 ERROR:", s3_url)
        return False, None

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] not in [HEADER_PATH, HEADER_URL, HEADER_S3_URL]:
            raise Exception(
                "Error with CSV file field: filename. Must be first field")


class ImageLoader(ParallelLoader.ParallelLoader):
    """
    **ApertureDB Image Loader.**

    This class is to be used in combination with a **generator** object,
    for example :class:`~aperturedb.ImageLoader.ImageGeneratorCSV`,
    which is a class that implements iterable inteface and generates "image data" elements.

    Example::

        image_data = {
            "properties":  properties,
            "constraints": constraints,
            "operations":  operations,
            "format":      format ("jpg", "png", etc),
            "img_blob":    (bytes),
        }
    """

    def __init__(self, db, dry_run=False):

        super().__init__(db, dry_run=dry_run)

        self.type = "image"

    def generate_batch(self, image_data):

        q = []
        blobs = []

        for data in image_data:

            ai = {
                "AddImage": {
                }
            }

            if "properties" in data:
                ai["AddImage"]["properties"] = data["properties"]
            if "constraints" in data:
                ai["AddImage"]["if_not_found"] = data["constraints"]
            if "operations" in data:
                ai["AddImage"]["operations"] = data["operations"]
            if "format" in data:
                ai["AddImage"]["format"] = data["format"]

            if "img_blob" not in data or len(data["img_blob"]) == 0:
                print("WARNING: Skipping empty image.")
                continue

            blobs.append(data["img_blob"])
            q.append(ai)

        return q, blobs



import time
import requests
import os
import logging
from os import path

import cv2
import numpy as np

from aperturedb import Parallelizer
from aperturedb import CSVParser

HEADER_PATH = "filename"
HEADER_URL  = "url"

logger = logging.getLogger(__name__)


class ImageDownloaderCSV(CSVParser.CSVParser):
    """**ApertureDB Image Downloader.**

    .. note::

        Expects a csv file with AT LEAST a "url" column, and
        optionally a "filename" field.
        If "filename" is not present, it is taken from the url.

    """

    def __init__(self, filename):

        self.has_filename = False

        super().__init__(filename)

    def __getitem__(self, idx):

        url = self.df.loc[idx, HEADER_URL]

        if self.has_filename:
            filename = self.df.loc[idx, HEADER_PATH]
        else:
            filename = self.url_to_filename(url)

        return url, filename

    def url_to_filename(self, url):

        filename = url.split("/")[-1]
        folder = "/tmp/images/"

        return folder + filename

    def validate(self):

        self.header = list(self.df.columns.values)

        if HEADER_URL not in self.header:
            raise Exception("Error with CSV file field: url. Must be a field")

        if HEADER_PATH in self.header:
            self.has_filename = True


class ImageDownloader(Parallelizer.Parallelizer):

    def __init__(self, n_download_retries=0, check_if_present=False):

        super().__init__()

        self.type = "image"

        self.check_img = check_if_present
        self.images_already_downloaded = 0
        self.n_download_retries = n_download_retries

    def check_if_image_is_ok(self, filename, url):

        if not os.path.exists(filename):
            return False

        try:
            a = cv2.imread(filename)
            if a.size <= 0:
                logger.warning("Image present but error reading it:", url)
                return False
        except Exception as e:
            logger.error("Image present but error decoding:", url)
            logger.exception(e)
            return False

        return True

    def download_image(self, url, filename):

        start = time.time()

        if self.check_img and self.check_if_image_is_ok(filename, url):
            self.images_already_downloaded += 1
            self.times_arr.append(time.time() - start)
            return

        folder = os.path.dirname(filename)
        if not os.path.exists(folder):
            os.makedirs(folder, exist_ok=True)

        retries = 0
        while True:
            imgdata = requests.get(url)
            if imgdata.ok:
                break
            else:
                if retries >= self.n_download_retries:
                    break
                logger.warning("Retrying object:", url)
                retries += 1
                time.sleep(2)

        if imgdata.ok:
            fd = open(filename, "wb")
            fd.write(imgdata.content)
            fd.close()

            try:
                a = cv2.imread(filename)
                if a.size <= 0:
                    logger.error("Downloaded image size error:", url)
                    os.remove(filename)
                    self.error_counter += 1
            except Exception as e:
                logger.error("Downloaded image cannot be decoded:", url)
                logger.exception(e)
                os.remove(filename)
                self.error_counter += 1
        else:
            logger.error("URL not found:", url)
            self.error_counter += 1

        self.times_arr.append(time.time() - start)

    def worker(self, thid, generator, start, end):

        for i in range(start, end):

            url, filename = generator[i]

            self.download_image(url, filename)

            if thid == 0 and self.stats:
                self.pb.update((i - start) / (end - start))

    def print_stats(self):

        times = np.array(self.times_arr)

        if len(times) <= 0:
            print("Error: No downloads.")
            return

        print("====== ApertureDB ImageDownloader Stats ======")
        if self.images_already_downloaded > 0:
            print("Images already present:", self.images_already_downloaded)

        print("Images downloaded:", len(times) -
              self.images_already_downloaded)
        print("Avg image time(s):", np.mean(times))
        print("Image time std:", np.std(times))
        print("Throughput (images/s)):",
              1 / np.mean(times) * self.numthreads)

        print("Total time(s):", self.total_actions_time)
        print("Overall throughput (img/s):",
              self.total_actions / self.total_actions_time)
        if self.error_counter > 0:
            print("Errors encountered:", self.error_counter)
        print("=============================================")



#
# The MIT License
#
# @copyright Copyright (c) 2017 Intel Corporation
# @copyright Copyright (c) 2021 ApertureData Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from . import queryMessage_pb2
import sys
import os
import socket
import struct
import time
import json
import ssl
import logging

from dataclasses import dataclass

logger = logging.getLogger(__name__)


PROTOCOL_VERSION = 1


@dataclass
class Session():

    session_token:      str
    refresh_token:      str
    session_token_ttl:  int
    refresh_token_ttl:  int
    session_started:    time.time = time.time()

    def valid(self) -> bool:
        session_age = time.time() - self.session_started

        if session_age > self.session_token_ttl:
            return False

        return True


class Connector(object):
    """
    .. _connector-label:

    **Class to facilitate connections with an instance of aperturedb**

    It lets the client execute any query as per the `Native Query specs <https://docs.aperturedata.io/>`_

    Args:
        str (host): Address of the host to connect to.
        int (port): Port to connect to.
        str (user): Username to specify while establishing a connection.
        str (password): Password to specify while connecting to the db.
        str (token): Token to use while connecting to the database.
        object (session):
        bool (use_ssl): Use SSL to encrypt communication with the database.
    """

    def __init__(self, host="localhost", port=55555,
                 user="", password="", token="",
                 session=None,
                 use_ssl=True):

        self.use_ssl = use_ssl

        self.host = host
        self.port = port

        self.connected = False
        self.last_response   = ''
        self.last_query_time = 0

        self.session = None

        self._connect()

        if session:
            self.session = session
        else:
            try:
                self._authenticate(user, password, token)
            except Exception as e:
                raise Exception("Authentication failed:", str(e))

    def __del__(self):

        self.conn.close()
        self.connected = False

    def _send_msg(self, data):

        sent_len = struct.pack('@I', len(data))  # send size first
        self.conn.send(sent_len)
        self.conn.send(data)

    def _recv_msg(self):

        recv_len = self.conn.recv(4)  # get message size

        recv_len = struct.unpack('@I', recv_len)[0]
        response = b''
        while len(response) < recv_len:
            packet = self.conn.recv(recv_len - len(response))
            if not packet:
                logger.error("Error receiving")
            response += packet

        return response

    def _authenticate(self, user, password="", token=""):

        query = [{
            "Authenticate": {
                "username": user,
            }
        }]

        if password:
            query[0]["Authenticate"]["password"] = password
        elif token:
            query[0]["Authenticate"]["token"] = token
        else:
            raise Exception(
                "Either password or token must be specified for authentication")

        response, _ = self._query(query)

        session_info = response[0]["Authenticate"]
        if session_info["status"] != 0:
            raise Exception(session_info["info"])

        self.session = Session(session_info["session_token"],
                               session_info["refresh_token"],
                               session_info["session_token_expires_in"],
                               session_info["refresh_token_expires_in"],
                               )

    def _check_session_status(self):

        if not self.session:
            return

        if not self.session.valid():
            self._refresh_token()

    def _refresh_token(self):

        query = [{
            "RefreshToken": {
                "refresh_token": self.session.refresh_token
            }
        }]

        response, _ = self._query(query, [])

        session_info = response[0]["RefreshToken"]
        if session_info["status"] != 0:
            raise Exception(session_info["info"])

        self.session = Session(session_info["session_token"],
                               session_info["refresh_token"],
                               session_info["session_token_expires_in"],
                               session_info["refresh_token_expires_in"],
                               )

    def _connect(self):

        self.conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.conn.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)

        # TCP_QUICKACK only supported in Linux 2.4.4+.
        # We use startswith for checking the platform following Python's
        # documentation:
        # https://docs.python.org/dev/library/sys.html#sys.platform
        if sys.platform.startswith('linux'):
            self.conn.setsockopt(socket.SOL_TCP, socket.TCP_QUICKACK, 1)

        self.conn.connect((self.host, self.port))

        # Handshake with server to negotiate protocol

        protocol = 2 if self.use_ssl else 1

        hello_msg = struct.pack('@II', PROTOCOL_VERSION, protocol)

        # Send desire protocol
        self._send_msg(hello_msg)

        # Receive response from server
        response = self._recv_msg()

        version, server_protocol = struct.unpack('@II', response)

        if version != PROTOCOL_VERSION:
            logger.warning("Protocol version differ from server")

        if server_protocol != protocol:
            self.conn.close()
            self.connected = False
            raise Exception(
                "Server did not accept protocol. Aborting Connection.")

        if self.use_ssl:

            # Server is ok with SSL, we switch over SSL.
            self.context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            self.context.check_hostname = False
            # TODO, we need to add support for local certificates
            # For now, we let the server send us the certificate
            self.context.verify_mode = ssl.VerifyMode.CERT_NONE
            self.conn = self.context.wrap_socket(self.conn)

        self.connected = True

    def _query(self, query, blob_array = []):

        # Check the query type
        if not isinstance(query, str):  # assumes json
            query_str = json.dumps(query)
        else:
            query_str = query

        if not self.connected:
            return "NOT CONNECTED"

        query_msg = queryMessage_pb2.queryMessage()
        # query has .json and .blobs
        query_msg.json = query_str

        # Set Auth token, only when not authenticated before
        if self.session:
            query_msg.token = self.session.session_token

        for blob in blob_array:
            query_msg.blobs.append(blob)

        # Serialize with protobuf and send
        data = query_msg.SerializeToString()
        self._send_msg(data)

        response = self._recv_msg()

        querRes = queryMessage_pb2.queryMessage()
        querRes.ParseFromString(response)

        response_blob_array = [b for b in querRes.blobs]

        self.last_response = json.loads(querRes.json)

        return (self.last_response, response_blob_array)

    def query(self, q, blobs=[]):

        self._check_session_status()

        try:
            start = time.time()
            self.response, self.blobs = self._query(q, blobs)
            self.last_query_time = time.time() - start
            return self.response, self.blobs
        except BaseException as e:
            print(e)
            raise ConnectionError("ApertureDB disconnected")

    def create_new_connection(self):

        return Connector(self.host, self.port, session=self.session)

    def get_last_response_str(self):

        return json.dumps(self.last_response, indent=4, sort_keys=False)

    def print_last_response(self):

        print(self.get_last_response_str())

    def get_last_query_time(self):

        return self.last_query_time

    def get_response(self):

        return self.response

    def get_blobs(self):

        return self.blobs

    # Return a flag on whether the last query succeeded:
    # i.e., all commands return status >= 0
    def last_query_ok(self):

        return self.check_status(self.response) >= 0

    def check_status(self, json_res):

        status = 0
        if (isinstance(self.response, dict)):
            if ("status" not in self.response):
                status = self.check_status(json_res[list(json_res.keys())[0]])
            else:
                status = json_res["status"]
        elif (isinstance(json_res, (tuple, list))):
            if ("status" not in json_res[0]):
                status = self.check_status(json_res[0])
            else:
                status = json_res[0]["status"]

        return status



import numpy as np

from test_Base import TestBase

from aperturedb import DescriptorDataCSV, Utils
from aperturedb.ImageDataCSV import ImageDataCSV
from aperturedb.ParallelLoader import ParallelLoader
from aperturedb.BlobDataCSV import BlobDataCSV
from aperturedb.EntityDataCSV import EntityDataCSV
from aperturedb.ConnectionDataCSV import ConnectionDataCSV
from aperturedb.DescriptorSetDataCSV import DescriptorSetDataCSV
from aperturedb.DescriptorDataCSV import DescriptorDataCSV

from aperturedb.BBoxDataCSV import BBoxDataCSV

import logging
logger = logging.getLogger(__name__)


def insert_data_from_csv(self: TestBase, in_csv_file, rec_count=-1):
    file_data_pair = {
        "./input/persons.adb.csv": EntityDataCSV,
        "./input/images.adb.csv": ImageDataCSV,
        "./input/connections-persons-images.adb.csv": ConnectionDataCSV,
        "./input/bboxes.adb.csv": BBoxDataCSV,
        "./input/blobs.adb.csv": BlobDataCSV,
        "./input/descriptorset.adb.csv": DescriptorSetDataCSV,
        "./input/setA.adb.csv": DescriptorDataCSV,
        "./input/setB.adb.csv": DescriptorDataCSV,
        "./input/s3_images.adb.csv": ImageDataCSV,
        "./input/http_images.adb.csv": ImageDataCSV,
        './input/bboxes-constraints.adb.csv': BBoxDataCSV
    }

    data = file_data_pair[in_csv_file](in_csv_file)
    if rec_count != -1:
        data = data[:rec_count]

    if self.stats:
        print("\n")

    loader = ParallelLoader(self.db)
    loader.ingest(data, batchsize=self.batchsize,
                  numthreads=self.numthreads,
                  stats=self.stats)
    self.assertEqual(loader.error_counter, 0)
    return data


class TestEntityLoader(TestBase):

    def test_Loader(self):

        self.db = self.create_connection()
        dbutils = Utils.Utils(self.db)
        classes = ["_Image", "_Descriptor", "Person", "_BoundingBox"]
        for c in classes:
            # Assert that we have a clean slate to begin with.
            self.assertEqual(dbutils.remove_entities(c), True)
            self.assertEqual(dbutils.count_entities(c), 0)

        results = [dbutils.create_entity_index(
            e, "id", "integer") for e in classes]
        self.assertTrue(all(results))

        # Insert Person Nodes
        data = insert_data_from_csv(self, "./input/persons.adb.csv")
        self.assertEqual(len(data), dbutils.count_entities("Person"))

        # Insert Images
        data = insert_data_from_csv(
            self, in_csv_file = "./input/images.adb.csv")
        self.assertEqual(len(data), dbutils.count_images())

        # Insert Connections
        data = insert_data_from_csv(
            self, in_csv_file = "./input/connections-persons-images.adb.csv")
        self.assertEqual(
            len(data), dbutils.count_connections("has_image"))

        # Insert BBoxes
        data = insert_data_from_csv(self, in_csv_file="./input/bboxes.adb.csv")
        self.assertEqual(len(data), dbutils.count_bboxes())

    def test_LoaderDescriptorsImages(self):

        self.db = self.create_connection()
        dbutils = Utils.Utils(self.db)
        # Insert Images, images may be there already, and that case should
        # be handled correctly by contraints
        data = insert_data_from_csv(
            self, in_csv_file = "./input/images.adb.csv")
        self.assertEqual(len(data), dbutils.count_images())

        # Insert DescriptorsSet
        data = insert_data_from_csv(
            self, in_csv_file = "./input/descriptorset.adb.csv")
        self.assertEqual(
            len(data), dbutils.count_entities("_DescriptorSet"))

        # Insert Descriptors

        sets = ["setA", "setB"]

        total_descriptors = 0
        for setname in sets:
            data = insert_data_from_csv(self,
                                        in_csv_file = "./input/" + setname + ".adb.csv")

            total_descriptors += len(data)
            self.assertEqual(total_descriptors,
                             dbutils.count_entities("_Descriptor"))

    def test_BlobLoader(self):
        self.db = self.create_connection()
        dbutils = Utils.Utils(self.db)

        # Assert that we have a clean slate to begin with.
        self.assertEqual(dbutils.remove_entities("_Blob"), True)
        self.assertEqual(dbutils.count_entities("_Blob"), 0)

        data = insert_data_from_csv(
            self, in_csv_file = "./input/blobs.adb.csv")
        self.assertEqual(len(data), dbutils.count_entities("_Blob"))

        query = [{
            "FindBlob": {
                "blobs": True,
                "results": {
                    "list": ["license"],
                    "limit": 1
                }
            }
        }]

        res, blob = self.db.query(query)
        self.assertEqual(len(blob), 1)

        arr = np.frombuffer(blob[0])
        self.assertEqual(arr[2], 3.3)

    def test_conditional_add(self):
        self.db = self.create_connection()
        dbutils = Utils.Utils(self.db)
        logger.debug(f"Cleaning existing data")

        self.assertTrue(dbutils.remove_entities("_BoundingBox"))
        self.assertTrue(dbutils.remove_entities("_Image"))
        # insert one of the images from image csv, for bboxes to refer to.
        images = insert_data_from_csv(
            self, in_csv_file="./input/images.adb.csv", rec_count=1)
        self.assertEqual(len(images), 1)

        # Insert BBoxes with repeated entries.
        # There is just one unique entry in the input csv.
        logger.debug(f"Inserting bounding box data")
        boxes = insert_data_from_csv(
            self, in_csv_file="./input/bboxes-constraints.adb.csv")
        self.assertEqual(3, len(boxes))
        self.assertEqual(1, dbutils.count_bboxes())


class TestURILoader(TestBase):

    def test_S3Loader(self):

        # Insert Images
        self.db = self.create_connection()

        dbutils = Utils.Utils(self.db)

        count_before = dbutils.count_images()
        data = insert_data_from_csv(
            self, in_csv_file = "./input/s3_images.adb.csv")
        self.assertEqual(len(data), dbutils.count_images() - count_before)

    def test_HttpImageLoader(self):
        self.db = self.create_connection()

        dbutils = Utils.Utils(self.db)

        count_before = dbutils.count_images()

        data = insert_data_from_csv(
            self, in_csv_file = "./input/http_images.adb.csv")
        self.assertEqual(len(data), dbutils.count_images() - count_before)



import os
import cv2
import numpy as np

from PIL import Image
from IPython.display import Video
from IPython.display import display as ds

DESTINATION_FOLDER = "results"


def check_folder(folder):
    if not os.path.exists(folder):
        os.makedirs(folder)


def display(images_array, save=False):

    for im in images_array:
        nparr = np.fromstring(im, np.uint8)
        image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        ds(Image.fromarray(image))

    if save:
        check_folder(DESTINATION_FOLDER)
        counter = 0
        for im in images_array:
            img_file = DESTINATION_FOLDER + '/res_' + str(counter) + '.jpg'
            counter += 1
            fd = open(img_file, 'wb')
            fd.write(im)
            fd.close()


def draw_bboxes(image, boxes=[], tags=[], save=False):

    nparr    = np.fromstring(image, np.uint8)
    cv_image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)

    # Draw a rectangle around the faces
    counter = 0
    for coords in boxes:
        left   = coords["x"]
        top    = coords["y"]
        right  = coords["x"] + coords["width"]
        bottom = coords["y"] + coords["height"]
        cv2.rectangle(cv_image, (left, top), (right, bottom), (0, 255, 0), 2)
        y = top - 15 if top - 15 > 15 else top + 15
        cv2.putText(cv_image, tags[counter], (left, y),
                    cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 255, 0), 2)
        counter += 1

    cv_image_rgb = cv2.cvtColor(cv_image, cv2.COLOR_BGR2RGB)
    ds(Image.fromarray(cv_image_rgb))

    if save:
        check_folder(DESTINATION_FOLDER)
        img_file = DESTINATION_FOLDER + '/res_bboxes.jpg'
        cv2.imwrite(img_file, cv_image)


def display_video_mp4(blob):

    check_folder(DESTINATION_FOLDER)

    name = DESTINATION_FOLDER + "/" + "video_tmp.mp4"
    fd = open(name, 'wb')
    fd.write(blob)
    fd.close()

    ds(Video(name, embed=True))
